==> ekeyd_1.1.1.orig.tar.gz: ekeyd-1.1.1/daemon/egd-linux.c <==
/* daemon/egd-linux.c
 *
 * Stand-alone EGD => Linux pool entropy transfer agent.
 *
 * Copyright 2009 Simtec Electronics.
 *
 * For licence terms refer to the COPYING file.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <stdbool.h>
#include <poll.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <syslog.h>

#include <linux/types.h>
#include <linux/random.h>
#include <sys/ioctl.h>

#include "daemonise.h"

#define DEFAULT_HOST "127.0.0.1"
#define DEFAULT_PORT 8888

static const char *pidfilename = NULL;
static int random_fd = 0;
static int egd_fd = -1;
static int shannons = 7;
static unsigned int retry_time = 0;
static struct sockaddr_in egd_addr;

struct pollfd poll_fds[2] = {
    { .events = POLLOUT },
    { .events = POLLIN }
};

#define RND_POLLFD 0
#define EGD_POLLFD 1

static int bytes_waiting = 0;

static void
usage(FILE *stream, char **argv)
{
    fprintf(stream,
            "%s: Usage:\n"                                              \
            "\n"                                                        \
            "%s [options]\n"                                            \
            "\n"                                                        \
            "Valid options are:\n"                                      \
            "\t-h\tDisplay this help.\n"                                \
            "\t-v\tDisplay the version of this program.\n"              \
            "\t-H <host>\tSet the host IP address to connect to.\n"     \
            "\t-p <postnr>\tSet the port number to connect to.\n"       \
            "\t-b <blocks>\tSet the number of 1024 bit blocks to request each time.\n" \
            "\t-s <S>\tSet the number of shannons per byte to S.\n" \
            "\t-D <pidf>\tDaemonise, writing PID to pidf.\n" \
            "\t-r <time>\tRetry connection every time seconds.\n",
            argv[0], argv[0]);
}

static bool
try_open_random(void)
{
    random_fd = open("/dev/random", O_RDWR);
    poll_fds[RND_POLLFD].fd = random_fd;
    return (random_fd != -1);
}

static void
set_egd_blocking(bool block)
{
    int flags;
    if ((flags = fcntl(egd_fd, F_GETFL)) == -1)
        perror("fcntl");
    flags &= ~(O_NONBLOCK);
    if (block == false)
        flags |= O_NONBLOCK;
    if (fcntl(egd_fd, F_SETFL, flags) == -1)
        perror("fcntl");
}

/** attempt to connect to an EGD server.
 *
 * Create a socket and connect it to an EGD server.
 *
 * @param rtime The time between retry attempts, 0 for no retrys.
 * @return true if the connection was sucessful, false if it was not.
 */
static bool
egd_connect(unsigned int rtime)
{
    int con_res = -1;
    int keepalive = 1;
    socklen_t keepalive_len = sizeof(keepalive);

    /* close fd if we already have one open */
    if (egd_fd >= 0) {
        close(egd_fd);
    }

    /* create socket */
    egd_fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (egd_fd == -1) {
        perror("socket");
        return false;
    }

    /* Set the keepalive option active */
   if (setsockopt(egd_fd, SOL_SOCKET, SO_KEEPALIVE, &keepalive, keepalive_len) < 0) {
       perror("setsockopt");
   }

    con_res = connect(egd_fd, (const struct sockaddr*)&egd_addr, sizeof(egd_addr));

    while ((rtime != 0) && (con_res != 0)) {
        sleep(retry_time);
        fprintf(stderr, "Retrying connection.\n");
        con_res = connect(egd_fd, (const struct sockaddr*)&egd_addr, sizeof(egd_addr));
    }

    if (con_res != 0) {
        perror("connect");
        close(egd_fd);
        egd_fd = -1;
        return false;
    }

    poll_fds[EGD_POLLFD].fd = egd_fd;

    /* ensure the random pool state is reset */
    bytes_waiting = 0;
    poll_fds[RND_POLLFD].events = POLLOUT;

    return true;
}


static void
issue_read_for_random(int nblocks)
{
    set_egd_blocking(true);
    bytes_waiting += nblocks * 128;
    while (nblocks--)
        if (write(egd_fd, (const void *)"\x02\x80", 2) != 2) {
            perror("write");
            bytes_waiting = 0;
            return;
        }
    set_egd_blocking(false);
}

static void
punt_entropy(void)
{
    static unsigned char buffer[128];
    struct rand_pool_info *rndpool;
    int bread;

    bread = read(egd_fd, buffer, 128);
    if (bread < 0) {
        perror("read");
        exit(1);
    }

    if (bread == 0) {
        /* EOF on the EGD socket is never good, reset the world */
        fprintf(stderr, "EGD Client Daemon reconnecting to server\n");
        syslog(LOG_ERR, "Reconnecting to EGD server");
        if (egd_connect(retry_time) == false) {
            exit(1);
        }
        return;
    }

    rndpool = alloca(sizeof(struct rand_pool_info) + bread);

    rndpool->entropy_count = bread * shannons;
    rndpool->buf_size = bread;
    memcpy(rndpool->buf, buffer, bread);

    if (ioctl(random_fd, RNDADDENTROPY, rndpool) == -1) {
        perror("ioctl");
        exit(1);
    }

    bytes_waiting -= bread;
    if (bytes_waiting == 0) {
        poll_fds[0].events = POLLOUT;
    }
}

int
main(int argc, char **argv)
{
    int opt;
    bool daemonise = false;
    int blocks = 4;

    /* setup default host address */
    egd_addr.sin_family = AF_INET;
    inet_pton(AF_INET, DEFAULT_HOST, &(egd_addr.sin_addr));
    egd_addr.sin_port = htons(DEFAULT_PORT);

    /* command line option parsing */
    while ((opt = getopt(argc, argv, "vhH:p:D:b:S:r:")) != -1) {
        switch (opt) {
        case 'v':
            printf("%s: Version 1.1\n", argv[0]);
            return 0;
            break;

        case 'h':
            usage(stdout, argv);
            return 0;
            break;

        case 'H':
            if (inet_pton(AF_INET, optarg, &(egd_addr.sin_addr)) != 1) {
                fprintf(stderr, "Host \"%s\" could not be parsed\n", optarg);
                return 1;
            }
            break;

        case 'p':
            egd_addr.sin_port = htons(strtoul(optarg, NULL, 10));
            break;

        case 'D':
            daemonise = true;
            pidfilename = optarg;
            break;

        case 'b':
            blocks = atoi(optarg);
            if (blocks < 1 || blocks > 4) {
                fprintf(stderr, "Reading %d blocks makes no sense\n", blocks);
                return 1;
            }
            break;

        case 'S':
            shannons = atoi(optarg);
            if (shannons < 1 || shannons > 8) {
                fprintf(stderr, "Claiming %d shannons per byte is pointless.\n", shannons);
                return 1;
            }
            break;

        case 'r':
            retry_time = strtoul(optarg, NULL, 10);
            break;

        default:
            usage(stderr, argv);
            return 1;
            break;
        }
    }

    if (try_open_random() == false) {
        fprintf(stderr, "Unable to open /dev/random for writing\n");
        return 1;
    }

    /* try initial connect to EGD server */
    egd_connect(0);
    if ((retry_time == 0) && (egd_fd == -1)) {
        fprintf(stderr, "Unable to connect to EGD server.\n");
        return 1;
    }

    if (daemonise)
        do_daemonise(pidfilename);

    /* now we are a daemon, start system logging */
    openlog("ekey-egd-linux", LOG_ODELAY, LOG_DAEMON);

    syslog(LOG_INFO, "Starting EGD Client Daemon");

    /* retry EGD server connection if necessary */
    if (egd_fd == -1)
        egd_connect(retry_time);

    /* ignore pipe signal */
    signal(SIGPIPE, SIG_IGN);

    /* Main poll for activity */
    while (poll(poll_fds, 2, -1) != -1) {

        if (poll_fds[RND_POLLFD].revents & (POLLERR | POLLHUP | POLLNVAL)) {
            syslog(LOG_INFO, "Linux random device poll error");
            break;
        }

        if (poll_fds[EGD_POLLFD].revents & (POLLERR | POLLHUP | POLLNVAL)) {
            if (retry_time != 0) {
                syslog(LOG_INFO, "EGD server poll error, reconnecting");
                sleep(retry_time);
                egd_connect(retry_time);
                continue;
            } else {
                syslog(LOG_ERR, "EGD server poll error");
                break;
            }
        }

        if (poll_fds[RND_POLLFD].revents & POLLOUT) {
            poll_fds[RND_POLLFD].events = 0;
            issue_read_for_random(blocks);
        }

        if (poll_fds[EGD_POLLFD].revents & POLLIN) {
            punt_entropy();
        }

    }

    if (egd_fd != -1)
        close(egd_fd);

    close(random_fd);

    syslog(LOG_INFO, "EGD Client Daemon Stopping");

    return 0;
}

==> ekeyd_1.1.2.orig.tar.gz: ekeyd-1.1.2/daemon/egd-linux.c <==
/* daemon/egd-linux.c
 *
 * Stand-alone EGD => Linux pool entropy transfer agent.
 *
 * Copyright 2009 Simtec Electronics.
 * Copyrithg 2010 Collabora Ltd
 *
 * For licence terms refer to the COPYING file.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <stdbool.h>
#include <poll.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <syslog.h>

#include <linux/types.h>
#include <linux/random.h>
#include <sys/ioctl.h>

#include "daemonise.h"

#define DEFAULT_HOST "127.0.0.1"
#define DEFAULT_PORT "8888"

static const char *pidfilename = NULL;
static int random_fd = 0;
static int egd_fd = -1;
static int shannons = 7;
static unsigned int retry_time = 0;
static char *host;
static char *port;

struct pollfd poll_fds[2] = {
    { .events = POLLOUT },
    { .events = POLLIN }
};

#define RND_POLLFD 0
#define EGD_POLLFD 1

static int bytes_waiting = 0;

static void
usage(FILE *stream, char **argv)
{
    fprintf(stream,
            "%s: Usage:\n"                                              \
            "\n"                                                        \
            "%s [options]\n"                                            \
            "\n"                                                        \
            "Valid options are:\n"                                      \
            "\t-h\tDisplay this help.\n"                                \
            "\t-v\tDisplay the version of this program.\n"              \
            "\t-H <host>\tSet the host IP address to connect to.\n"     \
            "\t-p <postnr>\tSet the port number to connect to.\n"       \
            "\t-b <blocks>\tSet the number of 1024 bit blocks to request each time.\n" \
            "\t-s <S>\tSet the number of shannons per byte to S.\n" \
            "\t-D <pidf>\tDaemonise, writing PID to pidf.\n" \
            "\t-r <time>\tRetry connection every time seconds.\n",
            argv[0], argv[0]);
}

static bool
try_open_random(void)
{
    random_fd = open("/dev/random", O_RDWR);
    poll_fds[RND_POLLFD].fd = random_fd;
    return (random_fd != -1);
}

static void
set_egd_blocking(bool block)
{
    int flags;
    if ((flags = fcntl(egd_fd, F_GETFL)) == -1)
        perror("fcntl");
    flags &= ~(O_NONBLOCK);
    if (block == false)
        flags |= O_NONBLOCK;
    if (fcntl(egd_fd, F_SETFL, flags) == -1)
        perror("fcntl");
}

/** attempt to connect to an EGD server.
 *
 * Create a socket and connect it to an EGD server.
 *
 * @param rtime The time between retry attempts, 0 for no retrys.
 * @return true if the connection was sucessful, false if it was not.
 */
static bool
egd_connect(unsigned int rtime)
{
    int con_res = -1;
    int keepalive = 1;
    socklen_t keepalive_len = sizeof(keepalive);
    struct addrinfo *addrs, *addrs_head;
    struct addrinfo hints;
    int r;

    /* Look up port and address */
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_flags = AI_ADDRCONFIG | AI_V4MAPPED;

    if ((r = getaddrinfo(host, port, &hints, &addrs)) != 0) {
        fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(r));
        return false;
    }

    addrs_head = addrs;
    do {
        while (addrs) {
            /* close fd if we already have one open */
            if (egd_fd >= 0) {
                close(egd_fd);
            }
            
            /* create socket */
            egd_fd = socket(addrs->ai_family, addrs->ai_socktype, addrs->ai_protocol);
            if (egd_fd == -1) {
                perror("socket");
            return false;
            }

            /* Set the keepalive option active */
            if (setsockopt(egd_fd, SOL_SOCKET, SO_KEEPALIVE, &keepalive, keepalive_len) < 0) {
                perror("setsockopt");
            }

            con_res = connect(egd_fd, addrs->ai_addr, addrs->ai_addrlen);
            if (con_res == 0) {
                rtime = 0; /* get out of outer loop */
                break;
            }
            addrs = addrs->ai_next;
        }
        if (rtime != 0) {
            sleep(rtime);
            fprintf(stderr, "Retrying connection.\n");
        }
        addrs = addrs_head;
    } while (rtime != 0);
    freeaddrinfo(addrs_head);

    /* Set the keepalive option active */
    if (setsockopt(egd_fd, SOL_SOCKET, SO_KEEPALIVE, &keepalive, keepalive_len) < 0) {
        perror("setsockopt");
    }

    if (con_res != 0) {
        perror("connect");
        close(egd_fd);
        egd_fd = -1;
        return false;
    }

    poll_fds[EGD_POLLFD].fd = egd_fd;

    /* ensure the random pool state is reset */
    bytes_waiting = 0;
    poll_fds[RND_POLLFD].events = POLLOUT;

    return true;
}


static void
issue_read_for_random(int nblocks)
{
    set_egd_blocking(true);
    bytes_waiting += nblocks * 128;
    while (nblocks--)
        if (write(egd_fd, (const void *)"\x02\x80", 2) != 2) {
            perror("write");
            bytes_waiting = 0;
            return;
        }
    set_egd_blocking(false);
}

static void
punt_entropy(void)
{
    static unsigned char buffer[128];
    struct rand_pool_info *rndpool;
    int bread;

    bread = read(egd_fd, buffer, 128);
    if (bread < 0) {
        perror("read");
        exit(1);
    }

    if (bread == 0) {
        /* EOF on the EGD socket is never good, reset the world */
        fprintf(stderr, "EGD Client Daemon reconnecting to server\n");
        syslog(LOG_ERR, "Reconnecting to EGD server");
        if (egd_connect(retry_time) == false) {
            exit(1);
        }
        return;
    }

    rndpool = alloca(sizeof(struct rand_pool_info) + bread);

    rndpool->entropy_count = bread * shannons;
    rndpool->buf_size = bread;
    memcpy(rndpool->buf, buffer, bread);

    if (ioctl(random_fd, RNDADDENTROPY, rndpool) == -1) {
        perror("ioctl");
        exit(1);
    }

    bytes_waiting -= bread;
    if (bytes_waiting == 0) {
        poll_fds[0].events = POLLOUT;
    }
}

int
main(int argc, char **argv)
{
    int opt;
    bool daemonise = false;
    int blocks = 4;

    /* setup default host address */
    host = DEFAULT_HOST;
    port = DEFAULT_PORT;

    /* command line option parsing */
    while ((opt = getopt(argc, argv, "vhH:p:D:b:S:r:")) != -1) {
        switch (opt) {
        case 'v':
            printf("%s: Version 1.1\n", argv[0]);
            return 0;
            break;

        case 'h':
            usage(stdout, argv);
            return 0;
            break;

        case 'H':
            host = strdup(optarg);
            break;

        case 'p':
            port = strdup(optarg);
            break;

        case 'D':
            daemonise = true;
            pidfilename = optarg;
            break;

        case 'b':
            blocks = atoi(optarg);
            if (blocks < 1 || blocks > 4) {
                fprintf(stderr, "Reading %d blocks makes no sense\n", blocks);
                return 1;
            }
            break;

        case 'S':
            shannons = atoi(optarg);
            if (shannons < 1 || shannons > 8) {
                fprintf(stderr, "Claiming %d shannons per byte is pointless.\n", shannons);
                return 1;
            }
            break;

        case 'r':
            retry_time = strtoul(optarg, NULL, 10);
            break;

        default:
            usage(stderr, argv);
            return 1;
            break;
        }
    }

    if (try_open_random() == false) {
        fprintf(stderr, "Unable to open /dev/random for writing\n");
        return 1;
    }

    /* try initial connect to EGD server */
    egd_connect(0);
    if ((retry_time == 0) && (egd_fd == -1)) {
        fprintf(stderr, "Unable to connect to EGD server.\n");
        return 1;
    }

    if (daemonise)
        do_daemonise(pidfilename);

    /* now we are a daemon, start system logging */
    openlog("ekey-egd-linux", LOG_ODELAY, LOG_DAEMON);

    syslog(LOG_INFO, "Starting EGD Client Daemon");

    /* retry EGD server connection if necessary */
    if (egd_fd == -1)
        egd_connect(retry_time);

    /* ignore pipe signal */
    signal(SIGPIPE, SIG_IGN);

    /* Main poll for activity */
    while (poll(poll_fds, 2, -1) != -1) {

        if (poll_fds[RND_POLLFD].revents & (POLLERR | POLLHUP | POLLNVAL)) {
            syslog(LOG_INFO, "Linux random device poll error");
            break;
        }

        if (poll_fds[EGD_POLLFD].revents & (POLLERR | POLLHUP | POLLNVAL)) {
            if (retry_time != 0) {
                syslog(LOG_INFO, "EGD server poll error, reconnecting");
                sleep(retry_time);
                egd_connect(retry_time);
                continue;
            } else {
                syslog(LOG_ERR, "EGD server poll error");
                break;
            }
        }

        if (poll_fds[RND_POLLFD].revents & POLLOUT) {
            poll_fds[RND_POLLFD].events = 0;
            issue_read_for_random(blocks);
        }

        if (poll_fds[EGD_POLLFD].revents & POLLIN) {
            punt_entropy();
        }

    }

    if (egd_fd != -1)
        close(egd_fd);

    close(random_fd);

    syslog(LOG_INFO, "EGD Client Daemon Stopping");

    return 0;
}
